@@ -154,6 +154,9 @@ struct _AnxSourceTrack {
   anx_int64_t end_granule;
   anx_int64_t current_granule;
   int eos;
+  anx_int64_t basegranule;/*< Base granule */
+  int preroll;            /*< Preroll */
+  int granuleshift;       /*< Granuleshift */
 };
 
 struct _AnxSource {
@@ -167,6 +170,9 @@ struct _AnxSource {
   double end_time;
   void * custom_data;
   double current_time;
+  int written_secondaries;
+  double duration;
+  long byte_length;
 };
 
 #ifdef __cplusplus
