@@ -30,15 +30,11 @@
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
-
-
 #include <time.h>
 #include <string.h>
 #include <limits.h>
 
 #include <stdlib.h>
-#define anx_malloc malloc
-#define anx_free free 
 
 #define MIN(a,b) ((a)<(b)?(a):(b))
 
@@ -46,6 +42,16 @@
 #include <annodex/annodex.h>
 
 /*#define DEBUG*/
+#define ANX_ANX_DEPRECATED 1
+
+#if ANX_ANX_DEPRECATED
+AnxImporter *
+anx_importer_init (int i)
+{
+  return NULL;
+}
+
+#else
 
 typedef struct _AnxAnxSource AnxAnxSource;
 typedef struct _AnxAnxTrack AnxAnxTrack;
@@ -156,19 +162,21 @@ read_stream (ANNODEX * anx, double timebase, char * utc, void * user_data)
 
 static int
 read_track (ANNODEX * anx, long serialno, char * id, char * content_type,
-		anx_int64_t granule_rate_n, anx_int64_t granule_rate_d,
-		int nr_header_packets, void * user_data)
+	    anx_int64_t granule_rate_n, anx_int64_t granule_rate_d,
+	    int nr_header_packets, void * user_data)
 {
   AnxAnxData * aad = (AnxAnxData *)user_data;
   AnxSource * media = aad->anx_media;
   AnxSourceTrack * track = NULL;
 
+  if (find_track (aad, serialno)) return 0;
+
+  if (!strncmp (content_type, "text/x-cmml", 12)) return 0;
+
 #ifdef DEBUG
   printf ("anxanx: read track (%ld) %s\n", serialno, content_type);
 #endif
 
-  if (!strncmp (content_type, "text/x-cmml", 12)) return 0;
-
   track = (AnxSourceTrack *) anx_malloc (sizeof (AnxSourceTrack));
   if (track == NULL) return -1;
 
@@ -193,6 +201,10 @@ read_track (ANNODEX * anx, long serialno, char * id, char * content_type,
     }
   }
 
+  track->basegranule = anx_track_get_basegranule (anx, serialno);
+  track->preroll = anx_track_get_preroll (anx, serialno);
+  track->granuleshift = anx_track_get_granuleshift (anx, serialno);
+
   if (add_track (aad, serialno, track) == NULL) return -1;
 
   media->tracks = anx_list_append (media->tracks, track);
@@ -319,6 +331,10 @@ anxanx_open (const char * path, const char * id, int ignore_media,
   aad = anx_malloc (sizeof (AnxAnxData));
   m->custom_data = aad;
 
+#ifdef DEBUG
+  printf ("anxanx_open: opening %s\n", path);
+#endif
+
   if ((anxr = anx_open ((char *)path, ANX_READ)) == NULL)
     return NULL;
 
@@ -353,6 +369,9 @@ anxanx_open (const char * path, const char * id, int ignore_media,
     while (!anx_ready (anxr) && (n = anx_read (anxr, 1024)) > 0);
   }
 
+#ifdef DEBUG
+  printf ("anxanx_open: got media %p\n", (void *)m);
+#endif
 
   return m;
 }
@@ -362,6 +381,10 @@ anxanx_read_update (AnxSource * media)
 {
   AnxAnxData * aad = (AnxAnxData *)media->custom_data;
 
+#ifdef DEBUG
+  printf ("anxanx_read_update: in\n");
+#endif
+
  do_read:
 
   while ((aad->media_packets == NULL) &&
@@ -520,3 +543,4 @@ anx_importer_init (int i)
   else
     return NULL;
 }
+#endif /* #if ANX_ANX_DEPRECATED */
