@@ -42,13 +42,8 @@
 
 #include <cmml.h>
 
-#define anx_malloc malloc
-#define anx_free free 
-
 #define MIN(a,b) ((a)<(b)?(a):(b))
 
-#define anx_strdup(s) ((s)==NULL?NULL:strdup(s))
-
 #define BUFSIZE 1024
 
 /*#define DEBUG*/
@@ -68,6 +63,33 @@ struct _AnxCMML_Data {
   double end_time;
 };
 
+static char *
+anxcmml_strdup (const char * s)
+{
+  char * ret;
+  if (!s) return NULL;
+  ret = anx_malloc (strlen (s) + 1);
+  return strcpy (ret, s);
+}
+
+static AnxMetaElement *
+anx_meta_from_cmml_meta (const CMML_MetaElement * meta)
+{
+  AnxMetaElement * new_meta;
+
+  if (meta == NULL) return NULL;
+
+  new_meta = (AnxMetaElement* ) anx_malloc (sizeof (AnxMetaElement));
+  new_meta->id = anxcmml_strdup (meta->id);
+  new_meta->lang = anxcmml_strdup (meta->lang);
+  new_meta->dir = anxcmml_strdup (meta->dir);
+  new_meta->name = anxcmml_strdup (meta->name);
+  new_meta->content = anxcmml_strdup (meta->content);
+  new_meta->scheme = anxcmml_strdup (meta->scheme);
+
+  return new_meta;
+}
+
 static AnxHead *
 anx_head_from_cmml_head (const CMML_Head * head)
 {
@@ -76,42 +98,44 @@ anx_head_from_cmml_head (const CMML_Head * head)
   h = anx_malloc (sizeof (AnxHead));
 
   /* id attribute of header */
-  h->head_id = anx_strdup(head->head_id);
+  h->head_id = anxcmml_strdup(head->head_id);
 
   /* language of header */
-  h->lang = anx_strdup(head->lang);
+  h->lang = anxcmml_strdup(head->lang);
 
   /* directionality of lang */
-  h->dir = anx_strdup(head->dir);
+  h->dir = anxcmml_strdup(head->dir);
 
   /* profile of header */
-  h->profile = anx_strdup(head->profile);
+  h->profile = anxcmml_strdup(head->profile);
 
   /* title element of header */
-  h->title = anx_strdup(head->title);
+  h->title = anxcmml_strdup(head->title);
 
   /* id attribute of title element */
-  h->title_id = anx_strdup(head->title_id);
+  h->title_id = anxcmml_strdup(head->title_id);
 
   /* language of title */
-  h->title_lang = anx_strdup(head->title_lang);
+  h->title_lang = anxcmml_strdup(head->title_lang);
 
   /* directionality of title_lang */
-  h->title_dir = anx_strdup(head->title_dir);
+  h->title_dir = anxcmml_strdup(head->title_dir);
 
   /* id attribute of base element */
-  h->base_id = anx_strdup(head->base_id);
+  h->base_id = anxcmml_strdup(head->base_id);
 
   /* href attribute of base element */
-  h->base_href = anx_strdup(head->base_href);
+  h->base_href = anxcmml_strdup(head->base_href);
 
 #if 0
   /* id attribute of stream element */
-  h->stream_id = anx_strdup(head->stream_id);
+  h->stream_id = anxcmml_strdup(head->stream_id);
 #endif
 
   /* list of meta elements */
-  h->meta = anx_list_clone ((AnxList*)head->meta);
+  h->meta =
+    anx_list_clone_with ((AnxList*)head->meta,
+                         (AnxCloneFunc)anx_meta_from_cmml_meta);
 
   return h;
 }
@@ -126,7 +150,8 @@ read_stream (CMML * cmml, const CMML_Stream * stream, void * user_data)
   char * id;
   char * content_type;
   char * src;
-  double timebase = 0.0;
+  double presentation_time = 0.0;
+  double basetime = 0.0;
   char * utc = NULL;
   double start_time = 0.0;
   double seek_offset = 0.0;
@@ -134,10 +159,11 @@ read_stream (CMML * cmml, const CMML_Stream * stream, void * user_data)
 
   if (ac->import_import) {
 
-    timebase = stream->timebase ? stream->timebase->t.sec : 0.0;
+    presentation_time = ac->start_time;
+    basetime = stream->timebase ? stream->timebase->t.sec : 0.0;
     utc = stream->utc ? stream->utc->tstr : NULL;
 
-    ac->import_stream (timebase, utc, ac->import_user_data);
+    ac->import_stream (presentation_time, basetime, utc, ac->import_user_data);
 
     for (media_node = stream->import; 
 	 media_node; 
@@ -146,10 +172,10 @@ read_stream (CMML * cmml, const CMML_Stream * stream, void * user_data)
       m = (CMML_ImportElement *) media_node->data;
 
       /* ignore granulerate for now, but might be used for raw importers */
-      granulerate = anx_strdup(m->granulerate);
-      id = anx_strdup (m->id);
-      content_type = anx_strdup (m->contenttype);
-      src = anx_strdup (m->src);
+      granulerate = anxcmml_strdup(m->granulerate);
+      id = anxcmml_strdup (m->id);
+      content_type = anxcmml_strdup (m->contenttype);
+      src = anxcmml_strdup (m->src);
       start_time = m->start_time ? m->start_time->t.sec : 0.0;
       /* end_time = m->end_time ? m->end_time->t.sec : 0.0; */
 
@@ -191,7 +217,7 @@ read_head (CMML * cmml, const CMML_Head * head, void * user_data)
     ac->import_head (h, ac->import_user_data);
   }
 
-  anx_free (h);
+  anx_head_free (h);
 
   return 0;
 }
@@ -204,29 +230,29 @@ anx_clip_from_cmml_clip (const CMML_Clip * clip)
   a = anx_malloc (sizeof (AnxClip));
 
   /* copy clip data across */
-  a->clip_id = anx_strdup (clip->clip_id);
-  a->lang = anx_strdup (clip->lang);
-  a->dir = anx_strdup (clip->dir);
-  a->track = anx_strdup (clip->track);
-  a->anchor_id = anx_strdup (clip->anchor_id);
-  a->anchor_lang = anx_strdup (clip->anchor_lang);
-  a->anchor_dir = anx_strdup (clip->anchor_dir);
-  a->anchor_class = anx_strdup (clip->anchor_class);
-  a->anchor_href = anx_strdup (clip->anchor_href);
-  a->anchor_text = anx_strdup (clip->anchor_text);
-  a->img_id = anx_strdup (clip->img_id);
-  a->img_lang = anx_strdup (clip->img_lang);
-  a->img_dir = anx_strdup (clip->img_dir);
-  a->img_src = anx_strdup (clip->img_src);
-  a->img_alt = anx_strdup (clip->img_alt);
-  a->desc_id = anx_strdup (clip->desc_id);
-  a->desc_lang = anx_strdup (clip->desc_lang);
-  a->desc_dir = anx_strdup (clip->desc_dir);
-  a->desc_text = anx_strdup (clip->desc_text);
+  a->clip_id = anxcmml_strdup (clip->clip_id);
+  a->lang = anxcmml_strdup (clip->lang);
+  a->dir = anxcmml_strdup (clip->dir);
+  a->track = anxcmml_strdup (clip->track);
+  a->anchor_id = anxcmml_strdup (clip->anchor_id);
+  a->anchor_lang = anxcmml_strdup (clip->anchor_lang);
+  a->anchor_dir = anxcmml_strdup (clip->anchor_dir);
+  a->anchor_class = anxcmml_strdup (clip->anchor_class);
+  a->anchor_href = anxcmml_strdup (clip->anchor_href);
+  a->anchor_text = anxcmml_strdup (clip->anchor_text);
+  a->img_id = anxcmml_strdup (clip->img_id);
+  a->img_lang = anxcmml_strdup (clip->img_lang);
+  a->img_dir = anxcmml_strdup (clip->img_dir);
+  a->img_src = anxcmml_strdup (clip->img_src);
+  a->img_alt = anxcmml_strdup (clip->img_alt);
+  a->desc_id = anxcmml_strdup (clip->desc_id);
+  a->desc_lang = anxcmml_strdup (clip->desc_lang);
+  a->desc_dir = anxcmml_strdup (clip->desc_dir);
+  a->desc_text = anxcmml_strdup (clip->desc_text);
 
   a->meta =
     anx_list_clone_with ((AnxList *)clip->meta,
-			 (AnxCloneFunc)anx_meta_element_clone);
+                         (AnxCloneFunc)anx_meta_from_cmml_meta);
 
   return a;
 }
@@ -235,10 +261,10 @@ static int
 read_clip (CMML * cmml, const CMML_Clip * clip, void * user_data)
 {
   AnxCMML_Data * ac = (AnxCMML_Data *)user_data;
-  AnxClip * a;
+  AnxClip * c;
 
   if (ac->import_clip) {
-    a = anx_clip_from_cmml_clip (clip);
+    c = anx_clip_from_cmml_clip (clip);
 
 #ifdef DEBUG
     {
@@ -248,21 +274,19 @@ read_clip (CMML * cmml, const CMML_Clip * clip, void * user_data)
       start = clip->start_time ? clip->start_time->t.sec : -1.0;
       end = clip->end_time ? clip->end_time->t.sec : -1.0;
 
-      anx_clip_snprint (buf, 1024, a, start, end);
+      anx_clip_snprint (buf, 1024, c, start, end);
       printf ("anxcmml::read_clip: %s\n", buf);
     }
 #endif
 
     if (clip->start_time != NULL) {
-      ac->import_clip (a, clip->start_time->t.sec,
-			 ac->import_user_data);
+      ac->import_clip (c, clip->start_time->t.sec, ac->import_user_data);
     }
     if (clip->end_time != NULL) {
-      ac->import_clip (NULL, clip->end_time->t.sec,
-			 ac->import_user_data);
+      ac->import_clip (NULL, clip->end_time->t.sec, ac->import_user_data);
     }
 
-    anx_free (a);
+    anx_clip_free (c);
   }
 
   return 0;
@@ -339,8 +363,11 @@ anxcmml_open (const char * path, const char * id, int ignore_media,
   m->eos = 1;
   m->start_time = 0;
   m->end_time = 0;
-  m->custom_data = NULL;
+  m->custom_data = ac;
   m->current_time = start_time;
+  m->written_secondaries = 1;
+  m->byte_length = 0;
+  m->duration = 0.0;
 
   return m;
 }
@@ -348,6 +375,9 @@ anxcmml_open (const char * path, const char * id, int ignore_media,
 static int
 anxcmml_close (AnxSource * media)
 {
+  AnxCMML_Data * ac = media->custom_data;
+
+  anx_free (ac);
   anx_free (media);
   return 0;
 }
