@@ -37,6 +37,7 @@
 
 #include <string.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 
@@ -58,9 +59,8 @@
 #include <windef.h>
 #define S_ISREG(x) TRUE
 #endif
- 
-
 
+static AnxList * importer_plugins = NULL;
 static AnxList * media_importers = NULL;
 
 int
@@ -116,7 +116,11 @@ anx_register_media_importer_lib (const char * pathname,
     importer_func = (AnxImporterInitFunc)
       dlsym_F (dlhandle, "anx_importer_init");
 
-    if (importer_func != NULL) {
+    if (importer_func == NULL) {
+      dlclose_F (dlhandle);
+    } else {
+      importer_plugins = anx_list_append (importer_plugins, dlhandle);
+      
       for (i = 0; (importer = importer_func(i)) != NULL; i++) {
 	if (anx_content_type_matches (importer->content_type,
 				      content_type_pattern))
@@ -124,7 +128,6 @@ anx_register_media_importer_lib (const char * pathname,
       }
     }
   }
-
 }
 
 static void
@@ -151,12 +154,28 @@ anx_register_all_media_importers_dir (char * dirname, char * content_type_patter
 	anx_register_media_importer_lib (buf, content_type_pattern);
     }
   }
+
+  closedir (dir);
+}
+
+static void
+unregister_plugins (void)
+{
+  AnxList * l;
+
+  for (l = importer_plugins; l; l = l->next) {
+    dlclose_F (l->data);
+  }
+
+  anx_list_free (importer_plugins);
+  anx_list_free (media_importers);
 }
 
 void
 anx_init_importers (char * content_type_pattern)
 {
   anx_register_all_media_importers_dir (IMPORTERS_DIR, content_type_pattern);
+  atexit (unregister_plugins);
 }
 
 static AnxImporter *
@@ -181,6 +200,7 @@ anx_media_open (ANNODEX * anx, char * filename, char * id,
 		AnxImportStream import_stream,
 		AnxImportHead import_head,
 		AnxImportClip import_clip,
+		AnxImportCMML import_cmml,
 		void * import_user_data,
 		AnxImportImport import_import,
 		void * import_import_user_data)
@@ -220,6 +240,7 @@ anx_media_open (ANNODEX * anx, char * filename, char * id,
   if ((flags & ANX_IGNORE_ANNO) == 0) {
     cb.import_head = import_head;
     cb.import_clip = import_clip;
+    cb.import_cmml = import_cmml;
   }
 
   cb.import_user_data = import_user_data;
@@ -271,6 +292,7 @@ anx_media_openfd (ANNODEX * anx, int fd, char * id, char * content_type,
 		  double start_time, double end_time, int flags,
 		  AnxImportHead import_head,
 		  AnxImportClip import_clip,
+		  AnxImportCMML import_cmml,
 		  void * import_user_data,
 		  AnxImportImport import_import,
 		  void * import_import_user_data)
@@ -363,7 +385,9 @@ anx_media_query_track(AnxSource * media, int track_index,
 		      char ** id, char ** content_type,
 		      long * nr_header_packets,
 		      anx_int64_t * granule_rate_n,
-		      anx_int64_t * granule_rate_d)
+		      anx_int64_t * granule_rate_d,
+		      anx_int64_t * basegranule,
+		      int * preroll, int * granuleshift)
 {
   AnxList * l;
   AnxSourceTrack * track;
@@ -378,6 +402,9 @@ anx_media_query_track(AnxSource * media, int track_index,
       *nr_header_packets = track->nr_header_packets;
       *granule_rate_n = track->granule_rate_n;
       *granule_rate_d = track->granule_rate_d;
+      *basegranule = track->basegranule;
+      *preroll = track->preroll;
+      *granuleshift = track->granuleshift;
       return 0;
     }
     i++;
@@ -447,6 +474,20 @@ anx_media_tell_time (AnxSource * media)
 #endif
 }
 
+long
+anx_media_get_byte_length (AnxSource * media)
+{
+  if (media == NULL) return 0;
+  return media->byte_length;
+}
+
+double
+anx_media_get_duration (AnxSource * media)
+{
+  if (media == NULL) return -1.0;
+  return media->duration;
+}
+
 int
 anx_media_eos (AnxSource * media)
 {
@@ -464,6 +505,13 @@ anx_media_current_track_eos (AnxSource * media)
   return (media->current_track->eos);
 }
 
+int
+anx_media_written_secondaries (AnxSource * media)
+{
+  if (media == NULL) return 0;
+  return media->written_secondaries;
+}
+
 #else /* ANX_CONFIG_WRITE */
 
 #include <annodex/anx_import.h>
