@@ -42,25 +42,36 @@
 #include <time.h>
 
 #include "anx_private.h"
+#include "anx_snprint.h" /* anx_strdup */
 
-/* #define DEBUG */
+/*#define DEBUG*/
 
 /* set to 0 or 1 */
 #define FLUSH_ALWAYS 0
 
-#define CMML_CONTENT_TYPE "text/x-cmml"
-#define EMPTY_CLIP "<clip/>"
+/* Tolerance for floating point rounding errors: used for portable
+ * floating point comparisons in anx_writer_sync()
+ */
+#define TOLERANCE 0.000000000001
 
 static char * cmml_content_type = CMML_CONTENT_TYPE;
 
+typedef enum {
+  ANX_DONT_FREE_AFTER = 0,
+  ANX_FREE_AFTER = 1
+} anx_free_after_t;
+
 static int
-anx_import_stream_default (double timebase, char * utc, void * user_data)
+anx_import_stream_default (double presentation_time, double basetime,
+			   char * utc, void * user_data)
 {
   ANNODEX * anx_writer = (ANNODEX *)user_data;
 
-  if (anx_writer->timebase == 0.0) {
-    anx_set_timebase (anx_writer, timebase);
-  }
+  if (anx_writer->presentation_time == 0.0)
+    anx_set_presentation_time (anx_writer, presentation_time);
+
+  if (anx_writer->basetime == 0.0)
+    anx_set_basetime (anx_writer, basetime);
 
   /* XXX: handle utc */
 
@@ -110,7 +121,18 @@ anx_import_import_default (double start_time, char * filename,
   anx_write_import (anx_writer, filename, id, content_type, seek_offset,
 		    seek_end, 0);
 
-  /* XXX: allow recursive imports to set timebase? */
+  /* XXX: allow recursive imports to set basetime? */
+
+  return 0;
+}
+
+static int
+anx_import_cmml_default (const char * cmml, long n, double time_offset,
+			 void * user_data)
+{
+  ANNODEX * anx_writer = (ANNODEX *)user_data;
+
+  anx_insert_cmml (anx_writer, time_offset, cmml, n);
 
   return 0;
 }
@@ -138,7 +160,8 @@ anx_import_import_recursive (double at_time, char * filename,
 static AnxWriterTrack *
 anx_writer_add_track (ANNODEX * annodex, AnxSource * media, int media_index,
 		      char * id, char * content_type, long nr_header_packets,
-		      anx_int64_t granule_rate_n, anx_int64_t granule_rate_d)
+		      anx_int64_t granule_rate_n, anx_int64_t granule_rate_d,
+		      anx_int64_t basegranule, int preroll, int granuleshift)
 {
   AnxWriter * writer = &annodex->x.writer;
   AnxWriterTrack * wsub;
@@ -152,15 +175,19 @@ anx_writer_add_track (ANNODEX * annodex, AnxSource * media, int media_index,
   track = (AnxTrack *)wsub;
 
   track->serialno = random ();
-  track->id = id;
-  track->content_type = content_type;
+  track->id = anx_strdup (id);
+  track->content_type = anx_strdup (content_type);
   track->nr_header_packets = nr_header_packets;
   track->granule_rate_n = granule_rate_n;
   track->granule_rate_d = granule_rate_d;
+  track->basegranule = basegranule;
+  track->preroll = preroll;
+  track->granuleshift = granuleshift;
 
   wsub->media = media;
   wsub->media_index = media_index;
 
+  wsub->written_fisbone = 0;
   wsub->written_bos = 0;
   wsub->written_eos = 0;
 
@@ -196,12 +223,23 @@ anx_writer_init_media (ANNODEX * annodex, AnxSource * media)
   char * id, * content_type;
   long nr_header_packets;
   anx_int64_t granule_rate, rate_interval;
+  anx_int64_t basegranule;
+  int preroll, granuleshift;
 
   if (media != NULL) {
+    double duration;
+
+    duration = anx_media_get_duration (media);
+    if (duration > annodex->duration)
+      annodex->duration = duration;
+
+    annodex->total_byte_length += anx_media_get_byte_length (media);
+
     for (i = 0; ; i++) {
       ret = anx_media_query_track (media, i, &id, &content_type,
 				   &nr_header_packets,
-				   &granule_rate, &rate_interval);
+				   &granule_rate, &rate_interval,
+				   &basegranule, &preroll, &granuleshift);
 
       if (ret < 0) break;
       
@@ -209,7 +247,8 @@ anx_writer_init_media (ANNODEX * annodex, AnxSource * media)
       printf ("anx_writer: adding track %d for %s\n", i, content_type);
 #endif
       anx_writer_add_track (annodex, media, i, id, content_type,
-			    nr_header_packets, granule_rate, rate_interval);
+			    nr_header_packets, granule_rate, rate_interval,
+			    basegranule, preroll, granuleshift);
     }
 
     writer->medias = anx_list_append (writer->medias, media);
@@ -227,6 +266,7 @@ anx_write_import (ANNODEX * annodex, char * filename,
   AnxImportStream import_stream = NULL;
   AnxImportHead import_head = NULL;
   AnxImportClip import_clip = NULL;
+  AnxImportCMML import_cmml = NULL;
   void * import_user_data = NULL;
 
   AnxImportImport import_import = NULL;
@@ -243,6 +283,7 @@ anx_write_import (ANNODEX * annodex, char * filename,
   if ((flags & ANX_IGNORE_ANNO) == 0) {
     import_head = writer->import_head_callback;
     import_clip = writer->import_clip_callback;
+    import_cmml = writer->import_cmml_callback;
     import_user_data = writer->import_user_data;
   }
 
@@ -258,7 +299,7 @@ anx_write_import (ANNODEX * annodex, char * filename,
 
   m = anx_media_open (annodex, filename, id, content_type,
 		      seek_offset, seek_end, flags,
-		      import_stream, import_head, import_clip,
+		      import_stream, import_head, import_clip, import_cmml,
 		      import_user_data,
 		      import_import, import_import_user_data);
 
@@ -282,6 +323,7 @@ int
 anx_write_set_anno_callbacks (ANNODEX * annodex,
 			      AnxImportHead import_head_callback,
 			      AnxImportClip import_clip_callback,
+			      AnxImportCMML import_cmml_callback,
 			      void * user_data)
 {
   AnxWriter * writer;
@@ -292,6 +334,7 @@ anx_write_set_anno_callbacks (ANNODEX * annodex,
 
   writer->import_head_callback = import_head_callback;
   writer->import_clip_callback = import_clip_callback;
+  writer->import_cmml_callback = import_cmml_callback;
   writer->import_user_data = user_data;
 
   return 0;
@@ -372,7 +415,7 @@ anx_writer_update_eos (ANNODEX * annodex)
 static long
 anx_packet_init (ANNODEX * annodex, unsigned char * buf, long n,
 		 long serialno, anx_int64_t granulepos, int bos, int eos,
-		 int flush)
+		 int flush, anx_free_after_t free_after)
 {
   /*AnxWriter * writer = &annodex->x.writer;*/
   ogg_packet op;
@@ -404,13 +447,156 @@ anx_packet_init (ANNODEX * annodex, unsigned char * buf, long n,
   ret = oggz_write_feed (annodex->oggz, &op, serialno, flush, NULL);
 #ifdef DEBUG
   if (ret < 0) {
-    printf ("oggz_write_feed returned %d\n", ret);
+    printf ("anx_packet_init: oggz_write_feed returned %d\n", ret);
   }
 #endif
 
+  if (free_after) anx_free (buf);
+
   return 0;
 }
 
+static long
+anx_init_fishead_packet (ANNODEX * annodex)
+{
+  unsigned char * fishead;
+  char * b;
+  anx_int64_t tn, td;
+
+  fishead = anx_malloc (FISHEAD_SIZE);
+  memset (fishead, 0, FISHEAD_SIZE);
+  b = (char *)fishead;
+
+  strncpy (b, FISHEAD_IDENTIFIER, 8);
+  b += 8;
+
+  *(ogg_int16_t *)b = _le_16 (FISHEAD_VERSION_MAJOR);
+  b += 2;
+
+  *(ogg_int16_t *)b = _le_16 (FISHEAD_VERSION_MINOR);
+  b += 2;
+
+  /* Presentation time */
+  tn = (anx_int64_t) (annodex->presentation_time * 1000);
+  td = 1000UL;
+
+  *(ogg_int64_t *)b = _le_64 (tn);
+  b += 8;
+
+  *(ogg_int64_t *)b = _le_64 (td);
+  b += 8;
+
+  /* Basetime */
+  tn = (anx_int64_t) (annodex->basetime * 1000);
+  td = 1000UL;
+
+  *(ogg_int64_t *)b = _le_64 (tn);
+  b += 8;
+
+  *(ogg_int64_t *)b = _le_64 (td);
+  b += 8;
+
+  /* UTC */
+  /* Leave as NUL bytes if not in use */
+
+  return anx_packet_init (annodex, fishead, FISHEAD_SIZE,
+			  annodex->mapping_serialno,
+			  0 /* granulepos */, 1 /* bos */, 0 /* eos */,
+			  OGGZ_FLUSH_AFTER, ANX_FREE_AFTER);
+}
+
+static long
+anx_init_fisbone_packet (ANNODEX * annodex, AnxTrack * track)
+{
+  unsigned char * buf;
+  char * b;
+  AnxParams * params = NULL;
+  size_t params_size, buf_size;
+
+  if (track == NULL) return -1;
+
+#ifdef DEBUG
+  printf ("anx_init_fisbone_packet: basegranule %lld\tpreroll %d\tgranuleshift %d\n",
+	  track->basegranule, track->preroll, track->granuleshift);
+#endif
+
+  params = anx_params_append (params, "Content-Type", track->content_type);
+  params = anx_params_append (params, "ID", track->id);
+
+  /* XXX: Non-C99 kludge */
+#if 0
+  params_size = anx_params_snprint (NULL, 0, params, ANX_PARAMS_HEADERS) + 1;
+
+  buf_size = FISBONE_SIZE + params_size;
+#else
+  buf_size = 1024;
+#endif
+
+  buf = anx_malloc (buf_size);
+  b = (char *)buf;
+
+  strncpy ((char *)b, FISBONE_IDENTIFIER, 8);
+  b += 8;
+
+  *(ogg_int32_t *)b = _le_32 (FISBONE_OFFSET);
+  b += 4;
+
+  *(ogg_int32_t *)b = _le_32 (track->serialno);
+  b += 4;
+
+  *(ogg_int32_t *)b = _le_32 (track->nr_header_packets);
+  b += 4;
+
+  *(ogg_int64_t *)b = _le_64 (track->granule_rate_n);
+  b += 8;
+
+  *(ogg_int64_t *)b = _le_64 (track->granule_rate_d);
+  b += 8;
+
+  *(ogg_int64_t *)b = _le_64 (track->basegranule);
+  b += 8;
+
+  *(ogg_int32_t *)b = _le_32 (track->preroll);
+  b += 4;
+
+  *(unsigned char *)b = (unsigned char)track->granuleshift;
+  b += 1;
+
+  b = (char *)buf + FISBONE_SIZE;
+
+  /* XXX: Non-C99 kludge */
+#if 0
+  anx_params_snprint (b, params_size, params, ANX_PARAMS_HEADERS);
+#else
+  params_size = anx_params_snprint (b, 1024 - FISBONE_SIZE, params,
+				    ANX_PARAMS_HEADERS);
+  buf_size = FISBONE_SIZE + params_size;
+#endif
+
+  anx_params_free (params);
+
+  return anx_packet_init (annodex, buf, buf_size, annodex->mapping_serialno,
+			  0 /* granulepos */, 0 /* bos */, 0 /* eos */,
+			  OGGZ_FLUSH_AFTER, ANX_FREE_AFTER);
+}
+
+static long
+anx_init_wsub_fisbone_packet (ANNODEX * annodex, AnxWriterTrack * wsub)
+{
+  AnxWriter * writer = &annodex->x.writer;
+  AnxTrack * track = (AnxTrack *)wsub;
+
+  if (wsub == NULL) return -1;
+
+  wsub->written_fisbone = 1;
+
+  writer->current_track = wsub;
+  /*writer->serialno = track->serialno;*/
+
+  return anx_init_fisbone_packet (annodex, track);
+}
+
+#if 0
 static long
 anx_init_mapping_packet (ANNODEX * annodex)
 {
@@ -431,7 +617,7 @@ anx_init_mapping_packet (ANNODEX * annodex)
   *(ogg_int16_t *)b = _le_16 (ANX_MAPPING_VERSION_MINOR);
   b += 2;
 
-  tn = (anx_int64_t) (annodex->timebase * 1000);
+  tn = (anx_int64_t) (annodex->basetime * 1000);
   td = 1000UL;
 
   *(ogg_int64_t *)b = _le_64 (tn);
@@ -443,11 +629,11 @@ anx_init_mapping_packet (ANNODEX * annodex)
   return anx_packet_init (annodex, mapping, ANX_MAPPING_SIZE,
 			  annodex->mapping_serialno,
 			  0 /* granulepos */, 1 /* bos */, 0 /* eos */,
-			  OGGZ_FLUSH_AFTER);
+			  OGGZ_FLUSH_AFTER, ANX_FREE_AFTER);
 }
 
 static long
-anx_init_track_bos_packet (ANNODEX * annodex, AnxTrack * track)
+anx_init_anxdata_packet (ANNODEX * annodex, AnxTrack * track)
 {
   unsigned char * buf;
   char * b;
@@ -463,7 +649,7 @@ anx_init_track_bos_packet (ANNODEX * annodex, AnxTrack * track)
 #if 0
   params_size = anx_params_snprint (NULL, 0, params, ANX_PARAMS_HEADERS) + 1;
 
-  buf_size = ANX_BOS_SIZE + params_size;
+  buf_size = ANX_DATA_SIZE + params_size;
 #else
   buf_size = 1024;
 #endif
@@ -471,9 +657,12 @@ anx_init_track_bos_packet (ANNODEX * annodex, AnxTrack * track)
   buf = anx_malloc (buf_size);
   b = (char *)buf;
 
-  strncpy ((char *)b, ANX_BOS_IDENTIFIER, 8);
+  strncpy ((char *)b, ANX_DATA_IDENTIFIER, 8);
   b += 8;
 
+  *(ogg_int32_t *)b = _le_32 (track->serialno);
+  b += 4;
+
   *(ogg_int64_t *)b = _le_64 (track->granule_rate_n);
   b += 8;
 
@@ -483,35 +672,82 @@ anx_init_track_bos_packet (ANNODEX * annodex, AnxTrack * track)
   *(ogg_int32_t *)b = _le_32 (track->nr_header_packets);
   b += 4;
 
-  b = (char *)buf + ANX_BOS_SIZE;
+  b = (char *)buf + ANX_DATA_SIZE;
 
   /* XXX: Non-C99 kludge */
 #if 0
   anx_params_snprint (b, params_size, params, ANX_PARAMS_HEADERS);
 #else
-  params_size = anx_params_snprint (b, 1024 - ANX_BOS_SIZE, params, ANX_PARAMS_HEADERS);
-  buf_size = ANX_BOS_SIZE + params_size;
+  params_size = anx_params_snprint (b, 1024 - ANX_DATA_SIZE, params, ANX_PARAMS_HEADERS);
+  buf_size = ANX_DATA_SIZE + params_size;
 #endif
 
-  return anx_packet_init (annodex, buf, buf_size, track->serialno,
-			  0 /* granulepos */, 1 /* bos */, 0 /* eos */,
-			  OGGZ_FLUSH_AFTER);
+  anx_params_free (params);
+
+  return anx_packet_init (annodex, buf, buf_size, annodex->mapping_serialno,
+			  0 /* granulepos */, 0 /* bos */, 0 /* eos */,
+			  OGGZ_FLUSH_AFTER, ANX_FREE_AFTER);
 }
 
 static long
-anx_init_wsub_bos_packet (ANNODEX * annodex, AnxWriterTrack * wsub)
+anx_init_wsub_anxdata_packet (ANNODEX * annodex, AnxWriterTrack * wsub)
 {
   AnxWriter * writer = &annodex->x.writer;
   AnxTrack * track = (AnxTrack *)wsub;
 
   if (wsub == NULL) return -1;
 
-  wsub->written_bos = 1;
+  wsub->written_anxdata = 1;
 
   writer->current_track = wsub;
   /*writer->serialno = track->serialno;*/
 
-  return anx_init_track_bos_packet (annodex, track);
+  return anx_init_anxdata_packet (annodex, track);
+}
+#endif
+
+static long
+anx_init_cmml_bos_packet (ANNODEX * annodex)
+{
+  unsigned char buf[CMML_HEADER_SIZE];
+  char * b = (char *)buf;
+
+  strncpy ((char *)b, CMML_IDENTIFIER, 4);
+  b[4] = '\0';
+  b[5] = '\0';
+  b[6] = '\0';
+  b[7] = '\0';
+  b += 8;
+
+  *(ogg_int16_t *)b = _le_16 (CMML_VERSION_MAJOR);
+  b += 2;
+
+  *(ogg_int16_t *)b = _le_16 (CMML_VERSION_MINOR);
+  b += 2;
+
+  *(ogg_int64_t *)b = _le_64 (CMML_GRANULE_RATE_N);
+  b += 8;
+
+  *(ogg_int64_t *)b = _le_64 (CMML_GRANULE_RATE_D);
+  b += 8;
+
+  *(unsigned char *)b = CMML_GRANULESHIFT;
+  b += 1;
+
+  return anx_packet_init (annodex, buf, CMML_HEADER_SIZE,
+			  annodex->cmml_serialno,
+			  0 /* granulepos */, 1 /* bos */, 0 /* eos */,
+			  OGGZ_FLUSH_AFTER, ANX_DONT_FREE_AFTER);
+}
+
+static long
+anx_init_cmml_preamble_packet (ANNODEX * annodex)
+{
+  return anx_packet_init (annodex, (unsigned char *)CMML_PREAMBLE,
+			  strlen (CMML_PREAMBLE),
+			  annodex->cmml_serialno,
+			  0 /* granulepos */, 0 /* bos */, 0 /* eos */,
+			  OGGZ_FLUSH_AFTER, ANX_DONT_FREE_AFTER);
 }
 
 static long
@@ -522,104 +758,120 @@ anx_init_head_packet (ANNODEX * annodex)
 
   input_n = (long)anx_head_snprint (NULL, 0, annodex->head);
 
-  if (input_n > writer->anno_buf_len) {
-    writer->anno_buf = realloc (writer->anno_buf, input_n);
-    writer->anno_buf_len = input_n;
+  if (input_n > writer->cmml_buf_len) {
+    writer->cmml_buf = realloc (writer->cmml_buf, input_n);
+    writer->cmml_buf_len = input_n;
   }
 
   input_n =
-    (long)anx_head_snprint (writer->anno_buf, writer->anno_buf_len,
+    (long)anx_head_snprint (writer->cmml_buf, writer->cmml_buf_len,
 			    annodex->head);
 
-  return anx_packet_init (annodex, (unsigned char *)writer->anno_buf, input_n,
-			  annodex->anno_serialno, 0 /* granulepos */,
+  return anx_packet_init (annodex, (unsigned char *)writer->cmml_buf, input_n,
+			  annodex->cmml_serialno, 0 /* granulepos */,
 			  0 /* bos */, 0 /* eos */,
-			  OGGZ_FLUSH_AFTER);
+			  OGGZ_FLUSH_AFTER, ANX_DONT_FREE_AFTER);
 }
 
-static long
-anx_init_clip_packet (ANNODEX * annodex)
+static anx_int64_t
+anx_writer_cmml_granulepos (ANNODEX * annodex)
 {
   AnxWriter * writer = &annodex->x.writer;
-  AnxTrack * track = &writer->anno_track;
+  AnxTrack * track = &writer->cmml_track;
   AnxCommand command;
   AnxCommandData data;
   double command_time;
-  anx_int64_t granulepos;
-  long input_n = 0;
+  anx_int64_t granule, granulepos;
 
   command = annodex->pending_command;
   if (command == NULL) return -1;
 
   data = anx_command_get_data (annodex, command);
   command_time = anx_command_get_time_offset (annodex, command);
-  granulepos = (anx_int64_t)(command_time * track->granule_rate_n /
-			     track->granule_rate_d);
+  granule = (anx_int64_t)(command_time * track->granule_rate_n /
+			  track->granule_rate_d);
+
+  granulepos =
+    (writer->cmml_prev_granule << CMML_GRANULESHIFT) |
+    (granule - writer->cmml_prev_granule);
+
+#ifdef DEBUG
+  printf ("anx_writer_cmml_granulepos: prev_granule %llx, granule %llx, granulepos %llx\n",
+	  writer->cmml_prev_granule, granule, granulepos);
+#endif
+
+  writer->cmml_prev_granule = granule;
+
+  return granulepos;
+}
+
+static long
+anx_init_clip_packet (ANNODEX * annodex)
+{
+  AnxWriter * writer = &annodex->x.writer;
+  AnxCommand command;
+  AnxCommandData data;
+  anx_int64_t granulepos;
+  long input_n = 0;
+
+  command = annodex->pending_command;
+  if (command == NULL) return -1;
+
+  granulepos = anx_writer_cmml_granulepos (annodex);
+
+  data = anx_command_get_data (annodex, command);
 
   if (data.clip != NULL) {
 
     input_n = (long)anx_clip_snprint (NULL, 0, data.clip, -1.0, -1.0);
 
-    if (input_n > writer->anno_buf_len) {
-      writer->anno_buf = realloc (writer->anno_buf, input_n);
-      writer->anno_buf_len = input_n;
+    if (input_n > writer->cmml_buf_len) {
+      writer->cmml_buf = realloc (writer->cmml_buf, input_n);
+      writer->cmml_buf_len = input_n;
     }
 
     input_n =
-      (long)anx_clip_snprint (writer->anno_buf, writer->anno_buf_len,
+      (long)anx_clip_snprint (writer->cmml_buf, writer->cmml_buf_len,
 			      data.clip, -1.0, -1.0);
   }
 
-  return anx_packet_init (annodex, (unsigned char *)writer->anno_buf, input_n,
-			  annodex->anno_serialno, granulepos,
+  return anx_packet_init (annodex, (unsigned char *)writer->cmml_buf, input_n,
+			  annodex->cmml_serialno, granulepos,
 			  0 /* bos */, 0 /* eos */,
-			  OGGZ_FLUSH_AFTER);
+			  OGGZ_FLUSH_AFTER, ANX_DONT_FREE_AFTER);
 }
 
 static long
 anx_init_clear_clip_packet (ANNODEX * annodex)
 {
   AnxWriter * writer = &annodex->x.writer;
-  AnxTrack * track = &writer->anno_track;
-  AnxCommand command;
-  double command_time;
   anx_int64_t granulepos;
   long input_n;
 
-  command = annodex->pending_command;
-  if (command == NULL) return -1;
+  granulepos = anx_writer_cmml_granulepos (annodex);
 
-  command_time = anx_command_get_time_offset (annodex, command);
-  granulepos = (anx_int64_t)(command_time * track->granule_rate_n /
-			     track->granule_rate_d);
+  input_n = (long) strlen (CMML_EMPTY_CLIP);
 
-  input_n = (long) strlen (EMPTY_CLIP);
-
-  if (input_n > writer->anno_buf_len) {
-    writer->anno_buf = realloc (writer->anno_buf, input_n);
-    writer->anno_buf_len = input_n;
+  if (input_n > writer->cmml_buf_len) {
+    writer->cmml_buf = realloc (writer->cmml_buf, input_n);
+    writer->cmml_buf_len = input_n;
   }
 
-  input_n = snprintf (writer->anno_buf, writer->anno_buf_len, EMPTY_CLIP);
+  input_n = snprintf (writer->cmml_buf, writer->cmml_buf_len, CMML_EMPTY_CLIP);
 
-  return anx_packet_init (annodex, (unsigned char *)writer->anno_buf, input_n,
-			  annodex->anno_serialno, granulepos,
+  return anx_packet_init (annodex, (unsigned char *)writer->cmml_buf, input_n,
+			  annodex->cmml_serialno, granulepos,
 			  0 /* bos */, 0 /* eos */,
-			  OGGZ_FLUSH_AFTER);
+			  OGGZ_FLUSH_AFTER, ANX_DONT_FREE_AFTER);
 }
 
 static long
-anx_writer_fill_media (ANNODEX * annodex)
+anx_writer_get_soonest_media (ANNODEX * annodex)
 {
   AnxWriter * writer = &annodex->x.writer;
   AnxSource * media, * soonest_media = NULL;
   AnxList * l;
-  AnxWriterTrack * wsub;
   double media_time, min_time = DBL_MAX;
-  int media_index;
-  long bound;
-  long read_len;
-  long input_n;
 
 #ifdef DEBUG
   int i = 0;
@@ -632,27 +884,21 @@ anx_writer_fill_media (ANNODEX * annodex)
     return -1;
   }
 
-  if (writer->need_bos) {
-#ifdef DEBUG
-    printf ("anx_writer_fill_media: need_bos\n");
-#endif
-    return 0;
-  }
-
-  bound = -1;
-
   /* update the current_media */
   for (l = writer->medias; l; l = l->next) {
     media = (AnxSource *)l->data;
-    media_time = anx_media_tell_time (media);
+    if (writer->secondary_headers_remaining == 0 ||
+	!anx_media_written_secondaries (media)) {
+      media_time = anx_media_tell_time (media);
 #ifdef DEBUG
-    printf ("anx_write: got media [%d] %p time %f %s\n", i++, media,
-	    media_time, anx_media_eos (media) ? "(eos)" : "");
+      printf ("anx_write: got media [%d] %p time %f %s\n", i++, media,
+	      media_time, anx_media_eos (media) ? "(eos)" : "");
 #endif
-    if (!anx_media_eos (media) &&
-	(media_time != -1 && media_time < min_time)) {
-      soonest_media = media;
-      min_time = media_time;
+      if (!anx_media_eos (media) &&
+	  (media_time != -1 && media_time < min_time)) {
+	soonest_media = media;
+	min_time = media_time;
+      }
     }
   }
 
@@ -660,6 +906,53 @@ anx_writer_fill_media (ANNODEX * annodex)
 
   writer->current_media = soonest_media;
 
+  return 0;
+}
+
+static long
+anx_writer_set_wsub (ANNODEX * annodex, AnxWriterTrack * wsub)
+{
+  AnxWriter * writer = &annodex->x.writer;
+  AnxSource * media, * soonest_media = NULL;
+  AnxList * l;
+  int media_index;
+
+  if (writer->media_buf_n != -1) {
+#ifdef DEBUG
+    printf ("anx_writer_fill_media: media_buf_n != -1\n");
+#endif
+    return -1;
+  }
+
+  /* update the current_media */
+  for (l = writer->medias; l; l = l->next) {
+    media = (AnxSource *)l->data;
+
+    /* Now grab the track details from the media */
+    media_index = anx_media_current_track (media);
+    if (wsub == anx_writer_find_track (annodex, media, media_index))
+      soonest_media = media;
+  }
+
+  if (soonest_media == NULL) return -1;
+
+  writer->current_media = soonest_media;
+  writer->current_track = wsub;
+  writer->media_serialno = wsub->sub.serialno;
+
+  return 0;
+}
+
+static long
+anx_writer_read_media (ANNODEX * annodex)
+{
+  AnxWriter * writer = &annodex->x.writer;
+  AnxWriterTrack * wsub;
+  int media_index;
+  long bound = -1;
+  long read_len;
+  long input_n;
+
   /* Read data from the current media, which in turn should set its
    * current track appropriately */
 
@@ -667,7 +960,7 @@ anx_writer_fill_media (ANNODEX * annodex)
 
   if (read_len > writer->media_buf_len) {
 #ifdef DEBUG
-    printf ("realloc media buf (%p) from %ld bytes to %ld bytes\n",
+    printf ("anx_writer_read_media: realloc media buf (%p) from %ld bytes to %ld bytes\n",
 	    writer->media_buf, writer->media_buf_len, read_len);
 #endif
     writer->media_buf = realloc (writer->media_buf, read_len);
@@ -680,13 +973,9 @@ anx_writer_fill_media (ANNODEX * annodex)
   writer->media_buf_n = input_n;
 
   /* Now grab the track details from the media */
-  media_index = anx_media_current_track (soonest_media);
+  media_index = anx_media_current_track (writer->current_media);
 
-#ifdef DEBUG
-  printf ("anx_write: media_index %d (soonest %p)\n", media_index,
-	  soonest_media);
-#endif
-  wsub = anx_writer_find_track (annodex, soonest_media, media_index);
+  wsub = anx_writer_find_track (annodex, writer->current_media, media_index);
 
   writer->current_track = wsub;
   writer->media_serialno = wsub->sub.serialno;
@@ -703,7 +992,7 @@ anx_init_media_packet (ANNODEX * annodex, int flush)
   AnxWriterTrack * wsub;
   long input_n = writer->media_buf_n;
   anx_int64_t granulepos;
-  int eos = 0;
+  int bos = 0, eos = 0;
 
   if (input_n == -1) return -1;
 
@@ -712,6 +1001,11 @@ anx_init_media_packet (ANNODEX * annodex, int flush)
   eos = anx_media_current_track_eos (writer->current_media);
   wsub = writer->current_track;
   if (wsub != NULL) {
+    writer->media_serialno = wsub->sub.serialno;
+    if (!wsub->written_bos) {
+      bos = 1;
+      wsub->written_bos = 1;
+    }
     if (wsub->written_eos) return -1;
     if (eos) {
       writer->current_track->written_eos = 1;
@@ -722,12 +1016,12 @@ anx_init_media_packet (ANNODEX * annodex, int flush)
     wsub->last_granulepos = granulepos;
   }
 
-  if (granulepos >= 0) flush = OGGZ_FLUSH_AFTER;
+  if (granulepos >= 0) flush |= OGGZ_FLUSH_AFTER;
 
   return anx_packet_init (annodex, writer->media_buf, input_n,
 			  writer->media_serialno,
 			  granulepos,
-			  0 /* bos */, eos, flush);
+			  bos, eos, flush, ANX_DONT_FREE_AFTER);
 }
 
 static long
@@ -737,21 +1031,23 @@ anx_init_eos_packet (ANNODEX * annodex, long serialno, anx_int64_t granulepos)
   printf ("anx_write: init eos for (%ld)\n", serialno);
 #endif
   return anx_packet_init (annodex, NULL, 0, serialno, granulepos,
-			  0 /* bos */, 1 /* eos */, 0 /* flush */);
+			  0 /* bos */, 1 /* eos */,
+			  OGGZ_FLUSH_BEFORE|OGGZ_FLUSH_AFTER /* flush */,
+			  ANX_DONT_FREE_AFTER);
 }
 
 static long
 anx_init_clip_eos_packet (ANNODEX * annodex)
 {
   AnxWriter * writer = &annodex->x.writer;
-  AnxTrack * track = &writer->anno_track;
+  AnxTrack * track = &writer->cmml_track;
   anx_int64_t granulepos;
 
   granulepos = (anx_int64_t)(anx_tell_time (annodex) *
 			     track->granule_rate_n /
 			     track->granule_rate_d);
 
-  return anx_init_eos_packet (annodex, annodex->anno_serialno, granulepos);
+  return anx_init_eos_packet (annodex, annodex->cmml_serialno, granulepos);
 }
 
 
@@ -782,43 +1078,113 @@ anx_writer_sync (ANNODEX * annodex)
   if (writer->done_media == 0) {
     /* Find the next packet and init it ... */
     
-    if (anx_writer_fill_media (annodex) == -1) {
+    if (writer->need_fishead) {
 #ifdef DEBUG
-      printf ("anx_writer_sync: fill_media returned -1\n");
+      printf ("anx_writer_sync: NEED fishead\n");
 #endif
-    }
-
-    if (writer->need_mapping) {
-      anx_init_mapping_packet (annodex);
-      writer->need_mapping = 0;
+      anx_init_fishead_packet (annodex);
+      writer->need_fishead = 0;
       made_packet = 1;
-    } else if (writer->need_anno_bos) {
-      /*anx_init_anno_bos_packet (annodex);*/
-      anx_init_track_bos_packet (annodex, &writer->anno_track);
-      writer->need_anno_bos = 0;
+    } else if (writer->need_cmml_bos) {
+#ifdef DEBUG
+      printf ("anx_writer_sync: NEED CMML BOS\n");
+#endif
+      anx_init_cmml_bos_packet (annodex);
+      writer->need_cmml_bos = 0;
       made_packet = 1;
     } else if (writer->need_bos) {
+#ifdef DEBUG
+      printf ("anx_writer_sync: NEED a BOS\n");
+#endif
       for (l = writer->tracks; l; l = l->next) {
 	wsub = (AnxWriterTrack *)l->data;
 	if (!wsub->written_bos) {
-	  anx_init_wsub_bos_packet (annodex, wsub);
+
+	  anx_writer_set_wsub (annodex, wsub);
+	  if (anx_writer_read_media (annodex) == -1) {
+#ifdef DEBUG
+	    printf ("anx_writer_sync: fill_media returned -1\n");
+#endif
+	  }
+
+	  anx_init_media_packet (annodex, OGGZ_FLUSH_BEFORE|OGGZ_FLUSH_AFTER);
+	  writer->secondary_headers_remaining--;
 	  made_packet = 1;
 	  break;
 	}
       }
 
       if (l == NULL) {
-	anx_init_eos_packet (annodex, annodex->mapping_serialno, 0);
+#ifdef DEBUG
+	if (!made_packet) printf ("anx_writer_sync: no BOS found\n");
+#endif
 	writer->need_bos = 0;
       }
+    } else if (writer->need_cmml_fisbone) {
+#ifdef DEBUG
+      printf ("anx_writer_sync: NEED CMML fisbone\n");
+#endif
+      anx_init_fisbone_packet (annodex, &writer->cmml_track);
+      writer->need_cmml_fisbone = 0;
+      made_packet = 1;
+    } else if (writer->need_fisbone) {
+#ifdef DEBUG
+      printf ("anx_writer_sync: NEED fisbone\n");
+#endif
+      for (l = writer->tracks; l; l = l->next) {
+	wsub = (AnxWriterTrack *)l->data;
+	if (!wsub->written_fisbone) {
+	  anx_init_wsub_fisbone_packet (annodex, wsub);
+	  writer->secondary_headers_remaining--;
+	  made_packet = 1;
+	  break;
+	}
+      }
+
+      if (l == NULL) {
+	writer->need_fisbone = 0;
+      }
+    } else if (writer->need_cmml_preamble) {
+#ifdef DEBUG
+      printf ("anx_writer_sync: NEED CMML preamble\n");
+#endif
+      anx_init_cmml_preamble_packet (annodex);
+      writer->need_cmml_preamble = 0;
+      made_packet = 1;
     } else if (writer->need_head) {
+#ifdef DEBUG
+      printf ("anx_writer_sync: NEED head\n");
+#endif
       anx_init_head_packet (annodex);
       writer->need_head = 0;
       made_packet = 1;
     } else if (writer->secondary_headers_remaining > 0) {
+#ifdef DEBUG
+      printf ("anx_write_sync: NEED %d secondaries remaing\n",
+	      writer->secondary_headers_remaining);
+#endif
+
+      if (anx_writer_get_soonest_media (annodex) == -1) {
+#ifdef DEBUG
+	printf ("anx_writer_sync: get_soonest_media returned -1\n");
+#endif
+      }
+
+      if (anx_writer_read_media (annodex) == -1) {
+#ifdef DEBUG
+	printf ("anx_writer_sync: read_media returned -1\n");
+#endif
+      }
+
       anx_init_media_packet (annodex, OGGZ_FLUSH_AFTER);
       writer->secondary_headers_remaining--;
       made_packet = 1;
+    } else if (writer->need_mapping_eos) {
+#ifdef DEBUG
+      printf ("anx_writer_sync: NEED mapping EOS\n");
+#endif
+      anx_init_eos_packet (annodex, annodex->mapping_serialno, 0);
+      writer->need_mapping_eos = 0;
     } else if (annodex->pending_command) {
 
       command_offset =
@@ -835,7 +1201,7 @@ anx_writer_sync (ANNODEX * annodex)
 #endif
       
       while (!made_packet &&
-	     ((current_offset > command_offset) ||
+	     ((current_offset > command_offset-TOLERANCE) ||
 	      (current_offset == 0.0 && command_offset == 0.0))) {
 
 	annodex->current_command = annodex->pending_command;
@@ -896,6 +1262,19 @@ anx_writer_sync (ANNODEX * annodex)
 
     /* Create a media packet */
     if (made_packet == 0) {
+
+      if (anx_writer_get_soonest_media (annodex) == -1) {
+#ifdef DEBUG
+	printf ("anx_writer_sync: get_soonest_media returned -1\n");
+#endif
+      }
+
+      if (anx_writer_read_media (annodex) == -1) {
+#ifdef DEBUG
+	printf ("anx_writer_sync: read_media returned -1\n");
+#endif
+      }
+
       anx_init_media_packet (annodex, 0 /* flush */);
 #ifdef DEBUG
       printf (".[%ld]", (long)anx_tell (annodex));
@@ -906,9 +1285,9 @@ anx_writer_sync (ANNODEX * annodex)
 #ifdef DEBUG
     printf ("anx_writer_sync: writer->done_media\n");
 #endif
-    if (!writer->done_anno_eos) {
+    if (!writer->done_cmml_eos) {
       anx_init_clip_eos_packet (annodex);
-      writer->done_anno_eos = 1;
+      writer->done_cmml_eos = 1;
       made_packet = 1;
     } else {
       for (l = writer->tracks; l; l = l->next) {
@@ -975,14 +1354,18 @@ anx_writer_init (ANNODEX * annodex)
   writer->current_media = NULL;
   writer->current_track = NULL;
 
-  writer->need_mapping = 1;
+  writer->need_fishead = 1;
   writer->need_bos = 1;
-  writer->need_anno_bos = 1;
+  writer->need_fisbone = 1;
+  writer->need_cmml_fisbone = 1;
+  writer->need_cmml_bos = 1;
+  writer->need_cmml_preamble = 1;
   writer->need_head = 1;
+  writer->need_mapping_eos = 1;
   writer->secondary_headers_remaining = 0;
 
   writer->done_media = 0;
-  writer->done_anno_eos = 0;
+  writer->done_cmml_eos = 0;
   writer->done_eos = 0;
 
   /* Initialise the commands list */
@@ -995,31 +1378,37 @@ anx_writer_init (ANNODEX * annodex)
   annodex->mapping_serialno = random ();
   annodex->mapping_serialno_valid = 1;
 
-  annodex->anno_serialno = random ();
-  annodex->anno_serialno_valid = 1;
+  annodex->cmml_serialno = random ();
+  annodex->cmml_serialno_valid = 1;
 
   writer->medias = NULL;
   writer->tracks = NULL;
 
-  writer->anno_track.serialno = annodex->anno_serialno;
-  writer->anno_track.id = NULL;
-  writer->anno_track.content_type = cmml_content_type;
-  writer->anno_track.nr_header_packets = 1;
-  writer->anno_track.granule_rate_n = 1000;
-  writer->anno_track.granule_rate_d = 1;
+  writer->cmml_track.serialno = annodex->cmml_serialno;
+  writer->cmml_track.id = NULL;
+  writer->cmml_track.content_type = cmml_content_type;
+  writer->cmml_track.nr_header_packets = 3;
+  writer->cmml_track.granule_rate_n = CMML_GRANULE_RATE_N;
+  writer->cmml_track.granule_rate_d = CMML_GRANULE_RATE_D;
+  writer->cmml_track.basegranule = 0;
+  writer->cmml_track.preroll = 0;
+  writer->cmml_track.granuleshift = CMML_GRANULESHIFT;
+
+  writer->cmml_prev_granule = 0;
 
   writer->eos = 0; /* end of stream */
 
   writer->import_stream_callback = anx_import_stream_default;
   writer->import_head_callback = anx_import_head_default;
   writer->import_clip_callback = anx_import_clip_default;
+  writer->import_cmml_callback = anx_import_cmml_default;
   writer->import_user_data = annodex;
 
   writer->import_import_callback = anx_import_import_default;
   writer->import_import_user_data = annodex;
 
-  writer->anno_buf = anx_malloc (102400);
-  writer->anno_buf_len = 102400;
+  writer->cmml_buf = anx_malloc (102400);
+  writer->cmml_buf_len = 102400;
 
   writer->media_buf = anx_malloc (102400);
   writer->media_buf_len = 102400;
@@ -1030,6 +1419,38 @@ anx_writer_init (ANNODEX * annodex)
   return annodex;
 }
 
+static AnxWriterTrack *
+anx_writer_track_free (AnxWriterTrack * track)
+{
+  anx_free (track->sub.content_type);
+  anx_free (track->sub.id);
+  anx_free (track);
+  return NULL;
+}
+
+int
+anx_writer_close (ANNODEX * annodex)
+{
+  AnxWriter * writer = &annodex->x.writer;
+  AnxSource * media;
+  AnxList * l;
+  
+  anx_list_free_with (writer->tracks, (AnxFreeFunc)anx_writer_track_free);
+
+  for (l = writer->medias; l; l = l->next) {
+    media = (AnxSource *)l->data;
+    anx_media_close (media);
+  }
+  anx_list_free (writer->medias);
+
+  anx_commands_delete (annodex);
+
+  anx_free (writer->cmml_buf);
+  anx_free (writer->media_buf);
+
+  return 0;
+}
+
 long
 anx_write_output (ANNODEX * annodex, unsigned char * buf, long n)
 {
@@ -1121,9 +1542,10 @@ anx_write_import (ANNODEX * annodex, char * filename,
 }
 
 int
-anx_write_set_anno_callbacks (ANNODEX * annodex,
-			       AnxImportHead import_head_callback,
-			       AnxImportClip import_clip_callback,
+anx_write_set_cmml_callbacks (ANNODEX * annodex,
+			      AnxImportHead import_head_callback,
+			      AnxImportClip import_clip_callback,
+			      AnxImportCMML import_cmml_callback,
 			       void * user_data)
 {
   return ANX_ERR_DISABLED;
