@@ -30,10 +30,13 @@
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
-/*
- * xtag.c
+/**
+ * \file xtag.c
  *
- * A trivial parser for xml-like tags
+ * XTag: A trivial parser for xml-like tags.
+ *
+ * API Version: 1.0
+ * Author: Conrad Parker, 2003
  */
 
 #if HAVE_CONFIG
@@ -50,6 +53,10 @@
 
 /*#define DEBUG*/
 
+#ifdef DEBUG
+#include <stdio.h>
+#endif
+
 #undef FALSE
 #undef TRUE
 
@@ -102,6 +109,8 @@ struct _XTagParser {
 #define X_SQUOTE      1<<4
 #define X_EQUAL       1<<5
 #define X_SLASH       1<<6
+#define X_DASH        1<<7
+#define X_QUESTION    1<<8
 
 static int
 xtag_cin (char c, int char_class)
@@ -127,6 +136,12 @@ xtag_cin (char c, int char_class)
   if (char_class & X_SLASH)
     if (c == '/') return TRUE;
 
+  if (char_class & X_DASH)
+    if (c == '-') return TRUE;
+
+  if (char_class & X_QUESTION)
+    if (c == '?') return TRUE;
+
   return FALSE;
 }
 
@@ -136,10 +151,10 @@ xtag_index (XTagParser * parser, int char_class)
   char * s;
   int i;
 
-  s = parser->start;
+  if (!parser->valid) return -1;
 
-  for (i = 0; s[i] && s != parser->end; i++) {
-    if (xtag_cin(s[i], char_class)) return i;
+  for (i = 0, s = parser->start; s != parser->end && *s; i++, s++) {
+    if (xtag_cin(*s, char_class)) return i;
   }
 
   return -1;
@@ -153,11 +168,9 @@ xtag_skip_over (XTagParser * parser, int char_class)
 
   if (!parser->valid) return;
 
-  s = (char *)parser->start;
-
-  for (i = 0; s[i] && s != parser->end; i++) {
-    if (!xtag_cin(s[i], char_class)) {
-      parser->start = &s[i];
+  for (i = 0, s = parser->start; s != parser->end && *s; i++, s++) {
+    if (!xtag_cin(*s, char_class)) {
+      parser->start = s;
       return;
     }
   }
@@ -171,27 +184,26 @@ xtag_skip_whitespace (XTagParser * parser)
   xtag_skip_over (parser, X_WHITESPACE);
 }
 
-#if 0
-static void
+static int
 xtag_skip_to (XTagParser * parser, int char_class)
 {
   char * s;
   int i;
 
-  if (!parser->valid) return;
+  if (!parser->valid) return FALSE;
 
-  s = (char *)parser->start;
-
-  for (i = 0; s[i] && s != parser->end; i++) {
-    if (xtag_cin(s[i], char_class)) {
-      parser->start = &s[i];
-      return;
+  for (i = 0, s = parser->start; s != parser->end && *s; i++, s++) {
+    if (xtag_cin(*s, char_class)) {
+      parser->start = s;
+      return TRUE;
     }
+#ifdef DEBUG
+    else printf ("xtag: skipping '%c'\n", s[i]);
+#endif
   }
 
-  return;
+  return FALSE;
 }
-#endif
 
 static char *
 xtag_slurp_to (XTagParser * parser, int good_end, int bad_end)
@@ -217,7 +229,7 @@ xtag_slurp_to (XTagParser * parser, int good_end, int bad_end)
 }
 
 static int
-xtag_assert_and_pass (XTagParser * parser, int char_class)
+xtag_pass (XTagParser * parser, int char_class, int assert)
 {
   char * s;
 
@@ -226,7 +238,7 @@ xtag_assert_and_pass (XTagParser * parser, int char_class)
   s = parser->start;
 
   if (!xtag_cin (s[0], char_class)) {
-    parser->valid = FALSE;
+    if (assert) parser->valid = FALSE;
     return FALSE;
   }
 
@@ -235,12 +247,24 @@ xtag_assert_and_pass (XTagParser * parser, int char_class)
   return TRUE;
 }
 
+static int
+xtag_check_and_pass (XTagParser * parser, int char_class)
+{
+  return xtag_pass (parser, char_class, FALSE);
+}
+
+static int
+xtag_assert_and_pass (XTagParser * parser, int char_class)
+{
+  return xtag_pass (parser, char_class, TRUE);
+}
+
 static char *
 xtag_slurp_quoted (XTagParser * parser)
 {
-  char * s, * ret;
+  char * s, * prev = NULL, * ret;
   int quote = X_DQUOTE; /* quote char to match on */
-  int xi;
+  int i;
 
   if (!parser->valid) return NULL;
 
@@ -254,16 +278,19 @@ xtag_slurp_quoted (XTagParser * parser)
 
   s = parser->start;
 
-  for (xi = 0; s[xi]; xi++) {
-    if (xtag_cin (s[xi], quote)) {
-      if (!(xi > 1 && s[xi-1] == '\\')) break;
+  for (i = 0, s = parser->start; s != parser->end && *s; i++, s++) {
+    if (xtag_cin (*s, quote)) {
+      if (!(i > 1 && *prev == '\\')) break;
     }
+    prev = s;
   }
 
-  ret = anx_malloc ((xi+1) * sizeof(char));
-  strncpy (ret, s, xi);
-  ret[xi] = '\0';
-  parser->start = &s[xi];
+  s = parser->start;
+
+  ret = anx_malloc ((i+1) * sizeof(char));
+  strncpy (ret, s, i);
+  ret[i] = '\0';
+  parser->start = &s[i];
 
   if (!xtag_assert_and_pass (parser, quote)) {
     anx_free (ret);
@@ -324,44 +351,131 @@ xtag_parse_attribute (XTagParser * parser)
 }
 
 static XTag *
-xtag_parse_tag (XTagParser * parser)
+xtag_parse_pcdata (XTagParser * parser)
 {
-  XTag * tag, * inner;
-  XAttribute * attr;
-  char * name;
+  XTag * tag;
   char * pcdata;
-  char * s;
 
   if (!parser->valid) return NULL;
 
-  if ((pcdata = xtag_slurp_to (parser, X_OPENTAG, X_NONE)) != NULL) {
-    tag = anx_malloc (sizeof (*tag));
-    tag->name = NULL;
-    tag->pcdata = pcdata;
-    tag->parent = parser->current_tag;
-    tag->attributes = NULL;
-    tag->children = NULL;
-    tag->current_child = NULL;
+  pcdata = xtag_slurp_to (parser, X_OPENTAG, X_NONE);
+  if (pcdata == NULL) return NULL;
 
-    return tag;
-  }
+  tag = anx_malloc (sizeof (*tag));
+  tag->name = NULL;
+  tag->pcdata = pcdata;
+  tag->parent = parser->current_tag;
+  tag->attributes = NULL;
+  tag->children = NULL;
+  tag->current_child = NULL;
+  
+  return tag;
+}
 
-  s = parser->start;
-
-  /* if this starts a close tag, return NULL and let the parent take it */
-  if (xtag_cin (s[0], X_OPENTAG) && xtag_cin (s[1], X_SLASH))
-    return NULL;
+static char *
+xtag_parse_name (XTagParser * parser)
+{
+  char * name;
 
   if (!xtag_assert_and_pass (parser, X_OPENTAG)) return NULL;
 
   name = xtag_slurp_to (parser, X_WHITESPACE | X_SLASH | X_CLOSETAG, X_NONE);
 
+  return name;
+}
+
+/* Forward declaration for tail-recursive skipping functions */
+static XTag * xtag_parse_tag (XTagParser * parser);
+
+static XTag *
+xtag_parse_directive (XTagParser * parser)
+{
+#ifdef DEBUG
+  printf ("xtag: processing directive\n");
+#endif
+  while (parser->start < parser->end &&
+	 xtag_skip_to (parser, X_QUESTION)) {
+    if (xtag_check_and_pass (parser, X_QUESTION) &&
+	xtag_check_and_pass (parser, X_CLOSETAG))
+      return xtag_parse_tag (parser);
+    }
+  /* Unterminated processing directive */
+  return NULL;
+}
+
+static XTag *
+xtag_parse_comment (XTagParser * parser)
+{
+#ifdef DEBUG
+  printf ("xtag: comment\n");
+#endif
+  while (parser->start < parser->end &&
+	 xtag_skip_to (parser, X_DASH)) {
+    if (xtag_check_and_pass (parser, X_DASH) &&
+	xtag_check_and_pass (parser, X_DASH) &&
+	xtag_check_and_pass (parser, X_CLOSETAG))
+      return xtag_parse_tag (parser);
+  }
+  /* Unterminated comment */
+  return NULL;
+}
+
+static XTag *
+xtag_parse_declaration (XTagParser * parser)
+{
+#ifdef DEBUG
+  printf ("xtag: non-comment declaration\n");
+#endif
+  while (parser->start < parser->end &&
+	 xtag_skip_to (parser, X_CLOSETAG)) {
+    if (xtag_check_and_pass (parser, X_CLOSETAG))
+      return xtag_parse_tag (parser);
+  }
+  /* Unterminated declaration */
+  return NULL;
+}
+
+static XTag *
+xtag_parse_tag (XTagParser * parser)
+{
+  XTag * tag, * inner;
+  XAttribute * attr;
+  char * name;
+  char * s;
+
+  if (!parser->valid) return NULL;
+
+  tag = xtag_parse_pcdata (parser);
+  if (tag != NULL) return tag;
+
+  s = parser->start;
+
+  /* Return to the parent if we're at the end of input */
+  if (s == parser->end) return NULL;
+
+  /* if this starts a close tag, return NULL and let the parent take it */
+  if (xtag_cin (s[0], X_OPENTAG) && xtag_cin (s[1], X_SLASH))
+    return NULL;
+
+  name = xtag_parse_name (parser);
   if (name == NULL) return NULL;
 
 #ifdef DEBUG
   printf ("<%s ...\n", name);
 #endif
 
+  /* Skip processing directives */
+  if (name[0] == '?') {
+    anx_free (name);
+    return xtag_parse_directive (parser);
+  } else if (!strcmp (name, "!--")) {
+    anx_free (name);
+    return xtag_parse_comment (parser);
+  } else if (name[0] == '!') {
+    anx_free (name);
+    return xtag_parse_declaration (parser);
+  }
+
   tag = anx_malloc (sizeof (*tag));
   tag->name = name;
   tag->pcdata = NULL;
@@ -396,8 +510,13 @@ xtag_parse_tag (XTagParser * parser)
     xtag_assert_and_pass (parser, X_OPENTAG);
     xtag_assert_and_pass (parser, X_SLASH);
     name = xtag_slurp_to (parser, X_WHITESPACE | X_CLOSETAG, X_NONE);
-    if (name) {
-      if (name && tag->name && strcmp (name, tag->name)) {
+    if (name == NULL) {
+#ifdef DEBUG
+      printf ("tag %s not closed\n", tag->name);
+#endif
+      parser->valid = FALSE;
+    } else {
+      if (tag->name && strcmp (name, tag->name)) {
 #ifdef DEBUG
 	printf ("got %s expected %s\n", name, tag->name);
 #endif
@@ -414,7 +533,6 @@ xtag_parse_tag (XTagParser * parser)
     xtag_assert_and_pass (parser, X_CLOSETAG);
   }
 
-
   return tag;
 }
 
@@ -470,6 +588,9 @@ xtag_new_parse (const char * s, int n)
   tag = xtag_parse_tag (&parser);
 
   if (!parser.valid) {
+#ifdef DEBUG
+    printf ("xtag_new_parse(): parser invalid\n");
+#endif
     xtag_free (tag);
     return NULL;
   }
@@ -630,8 +751,10 @@ xtag_snprints (char * buf, int n, ...)
     len = (int) strlen (s);
 
     if ((to_copy = MIN (n, len)) > 0) {
-      memcpy (buf, s, to_copy);
-      buf += to_copy;
+      if (buf) {
+	memcpy (buf, s, to_copy);
+	buf += to_copy;
+      }
       n -= to_copy;
     }
 
@@ -652,10 +775,13 @@ xtag_snprint (char * buf, int n, XTag * xtag)
   XTag * child;
 
 #define FORWARD(N) \
-  buf += MIN (n, N); \
-  n = MAX (n-N, 0);  \
+  if (buf) buf += MIN (n, N); \
+  n = MAX (n-N, 0);           \
   written += N;
 
+  /* If a NULL buffer is passed in, ensure it is not written to */
+  if (buf == NULL) n = 0;
+
   if (xtag == NULL) {
     if (n > 0) buf[0] = '\0';
     return 0;
